package fmcclient

// https://www.cisco.com/c/en/us/td/docs/security/firepower/620/api/REST/Firepower_Management_Center_REST_API_Quick_Start_Guide_620/Objects_in_the_REST_API.html#reference_jpj_jqc_bcb

import (
	"bytes"
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"log"
	"net/http"
	"net/url"
	"sync"
	"time"
)

var EOF = errors.New("EOF")

const MaxAuthTokenMinutes = 60
const MaxAuthTokenRefreshes = 3
const MinutesBeforeRefreshDeadline = 3

func NewAuthTokenExpiry(time.Time) time.Time {
	return time.Now().Add(MaxAuthTokenMinutes * time.Minute)
}

func withinAuthTokenRefreshWindow(expiry time.Time) bool {
	return time.Duration(0) < time.Until(expiry) &&
		time.Until(expiry) <= MinutesBeforeRefreshDeadline*time.Minute
}

func authTokenRefreshDeadlineHasPassed(
	expiry time.Time,
) bool {
	return time.Until(expiry) <= time.Duration(0)
}

func maxAuthTokenRefreshesHasBeenReached(cnt int) bool {
	return cnt >= MaxAuthTokenRefreshes
}

const (
	FMCStatusCodeOK                  int = 200
	FMCStatusCodeCreated             int = 201
	FMCStatusCodeAccepted            int = 202
	FMCStatusCodeNoContent           int = 204
	FMCStatusCodeBadRequest          int = 400
	FMCStatusCodeNotFound            int = 404
	FMCStatusCodeMethodNotAllowed    int = 405
	FMCStatusCodeUnprocessableEntity int = 422
	FMCStatusCodeTooManyRequests     int = 429
)

func newFMCError(
	resp *http.Response,
	host,
	msg string,
) FMCError {
	fmcErr := &FMCErrorResponse{
		Error: FMCError{
			Host:           host,
			HttpStatus:     http.StatusText(resp.StatusCode),
			HttpStatusCode: resp.StatusCode,
			Err:            msg,
		},
	}
	defer resp.Body.Close()
	body, err := io.ReadAll(resp.Body)
	if err != nil {
		return fmcErr.Error
	}
	json.Unmarshal(body, &fmcErr)
	return fmcErr.Error
}

type FMCError struct {
	Host           string
	HttpStatus     string
	HttpStatusCode int
	Category       string
	Severity       string
	Code           string
	Details        string
	Messages       []map[string]string
	Err            string
}

func (e FMCError) Error() string {
	status := http.StatusText(e.HttpStatusCode)
	if status != "" {
		status = " " + status
	}
	return fmt.Sprintf(
		"%s at FMC host %s: http status %d%s; category: %s; severity: %s; code: %s; details: %s; messages: %s",
		e.Err,
		e.Host,
		e.HttpStatusCode,
		status,
		e.Category,
		e.Severity,
		e.Code,
		e.Details,
		e.Messages,
	)
}

type FMCErrorResponse struct {
	Error FMCError
}

type Pager interface {
	Next() []interface{}
}

// Retrieve n items in the range [m*n, m*n + n - 1],
// inclusive
type pager struct {
	fmc *FMC
	url string
	m   uint64
	n   uint64
	eof bool
}

func (p *pager) next(ptr interface{}) error {
	if p.eof {
		return EOF
	}
	q := fmt.Sprintf(
		"?expanded=true&limit=%d&offset=%d",
		p.n,
		p.m*p.n,
	)
	resp, err := p.fmc.get(p.url + q)
	if resp.StatusCode != FMCStatusCodeOK {
		return newFMCError(resp, p.fmc.host, err.Error())
	}
	defer resp.Body.Close()
	body, err := io.ReadAll(resp.Body)
	if err != nil {
		return fmt.Errorf("read response body: %v", err)
	}
	//log.Printf("DEBUG: %s\n", body)
	head := &fmcResponseBase{}
	err = json.Unmarshal(body, head)
	if err != nil {
		return err
	}
	if (p.m*p.n + p.n) >= head.Paging.Count {
		p.eof = true
	}
	err = json.Unmarshal(body, ptr)
	if err != nil {
		log.Printf("DEBUG: %s\n", body)
		return err
	}
	p.m++
	return nil
}

/* From https://www.cisco.com/c/en/us/td/docs/security/firepower/620/api/REST/Firepower_Management_Center_REST_API_Quick_Start_Guide_620/Objects_in_the_REST_API.html:
> The REST API will serve only 25 results per page. This can
> be increased up to 1000 using the limit query parameter.*/
func newPager(url string, f *FMC, pageSize uint64) pager {
	if pageSize == 0 {
		pageSize = 25
	}
	if pageSize > 1000 {
		pageSize = 1000
	}
	return pager{
		fmc: f,
		url: url,
		m:   0,
		n:   pageSize,
	}
}

type FMCLinks struct {
	Self string `json:"self"`
}

type fmcResponseBase struct {
	Links  FMCLinks `json:"links"`
	Paging struct {
		Count  uint64 `json:"count"`
		Limit  uint64 `json:"limit"`
		Offset uint64 `json:"offset"`
		Pages  uint64 `json:"pages"`
	} `json:"paging"`
}

type FMCAccessPolicy struct {
	Links FMCLinks `json:"links"`
	Id    string   `json:"id"`
	Name  string   `json:"name"`
	Type  string   `json:"type"`
}

type fmcAccessPolicies struct {
	Items []*FMCAccessPolicy `json:"items"`
}

type fmcAccessPoliciesPager struct {
	pager
}

func (p *fmcAccessPoliciesPager) Next() (
	out []*FMCAccessPolicy,
	err error,
) {
	resp := &fmcAccessPolicies{}
	err = p.next(resp)
	if err == EOF {
		return
	}
	if err != nil {
		err = fmt.Errorf("fmc access policies pager: next: %v", err)
		return
	}
	out = resp.Items
	return
}

type FMCAccessPolicyRuleActionType string

const (
	FMCAccessPolicyRuleActionAllow   FMCAccessPolicyRuleActionType = "ALLOW"
	FMCAccessPolicyRuleActionTrust   FMCAccessPolicyRuleActionType = "TRUST"
	FMCAccessPolicyRuleActionMonitor FMCAccessPolicyRuleActionType = "MONITOR"
	FMCAccessPolicyRuleActionBlock   FMCAccessPolicyRuleActionType = "BLOCK"
	//FMCAccessPolicyRuleActionBlockWithReset FMCAccessPolicyRuleActionType = "BLOCK_WITH_RESET" ?
	//FMCAccessPolicyRuleActionInteractiveBlock FMCAccessPolicyRuleActionType = "INTERACTIVE_BLOCK" ?
	//FMCAccessPolicyRuleActionInteractiveBlockWithReset FMCAccessPolicyRuleActionType = "INTERACTIVE_BLOCK_WITH_RESET" ?
)

type FMCAccessPolicyRuleComment struct {
	Comment string `json:"comment"`
	Date    string `json:"date"`
	User    struct {
		Name string `json:"name"`
		Type string `json:"type"`
	} `json:"user"`
}

type FMCSecurityZoneObject struct {
	Id   string `json:"id"`
	Name string `json:"name,omitempty"`
	Type string `json:"type,omitempty"`
}

type FMCSyslogConfigObject struct {
	Id   string `json:"id"`
	Name string `json:"name,omitempty"`
	Type string `json:"type,omitempty"`
}

type FMCAccessPolicyRuleMetadata struct {
	Category  string `json:"category,omitempty"`
	Section   string `json:"section,omitempty"`
	RuleIndex uint16 `json:"ruleIndex,omitempty"`
}

type FMCIPSPolicy struct {
	Id         string `json:"id"`
	Name       string `json:"name,omitempty"`
	Type       string `json:"type,omitempty"`
	InlineDrop uint64 `json:"inlineDrop,omitempty"`
}

type FMCVariableSet struct {
	Id string `json:"id"`
}

type FMCAccessPolicyRuleZones struct {
	Objects []*FMCSecurityZoneObject `json:"objects"`
}

type FMCAccessPolicyRuleNetworks struct {
	Objects  []*FMCNetworkObjectRef `json:"objects,omitempty"`
	Literals []*FMCNetworkLiteral   `json:"literals,omitempty"`
}

type FMCPortObjectRef struct {
	Id          string `json:"id,omitempty"`
	Name        string `json:"name,omitempty"`
	Protocol    string `json:"protocol,omitempty"`
	Type        string `json:"type,omitempty"` // ProtocolPortObject|PortObjectGroup
	Overridable bool   `json:"overridable,omitempty"`
}

// Protocol -9999 represents "All"
type FMCPortLiteral struct {
	Port     string `json:"port,omitempty"`
	Protocol string `json:"protocol"` // 0-255,-9999
	Type     string `json:"type"`
	ICMPType string `json:"icmpType,omitempty"`
}

type FMCAccessPolicyRulePorts struct {
	Objects  []*FMCPortObjectRef `json:"objects,omitempty"`
	Literals []*FMCPortLiteral   `json:"literals,omitempty"`
}

type FMCAccessPolicyRule struct {
	Id                  string                        `json:"id,omitempty"`
	Name                string                        `json:"name"`
	Type                string                        `json:"type,omitempty"`
	Metadata            *FMCAccessPolicyRuleMetadata  `json:"metadata,omitempty"`
	Action              FMCAccessPolicyRuleActionType `json:"action"`
	Enabled             bool                          `json:"enabled"`
	EnableSyslog        bool                          `json:"enableSyslog"`
	SendEventsToFMC     bool                          `json:"sendEventsToFMC"`
	IPSPolicy           *FMCIPSPolicy                 `json:"ipsPolicy,omitempty"`
	VariableSet         *FMCVariableSet               `json:"variableSet,omitempty"`
	LogBegin            bool                          `json:"logBegin"`
	LogEnd              bool                          `json:"logEnd"`
	LogFiles            bool                          `json:"logFiles"`
	SourceZones         *FMCAccessPolicyRuleZones     `json:"sourceZones"`
	DestinationZones    *FMCAccessPolicyRuleZones     `json:"destinationZones"`
	SourceNetworks      *FMCAccessPolicyRuleNetworks  `json:"sourceNetworks"`
	DestinationNetworks *FMCAccessPolicyRuleNetworks  `json:"destinationNetworks"`
	DestinationPorts    *FMCAccessPolicyRulePorts     `json:"destinationPorts,omitempty"`
	CommentHistoryList  []*FMCAccessPolicyRuleComment `json:"commentHistoryList,omitempty"`
}

type fmcAccessPolicyRules struct {
	Items []*FMCAccessPolicyRule `json:"items"`
}

type fmcAccessPolicyRulesPager struct {
	pager
}

func (p *fmcAccessPolicyRulesPager) Next() (
	out []*FMCAccessPolicyRule,
	err error,
) {
	resp := &fmcAccessPolicyRules{}
	err = p.next(resp)
	if err == EOF {
		return
	}
	if err != nil {
		err = fmt.Errorf("fmc access policy rules pager: next: page=%d: %v", p.m, err)
		return
	}
	out = resp.Items
	return
}

type FMCNetworkObjectRef struct {
	Id          string `json:"id,omitempty"`
	Name        string `json:"name,omitempty"`
	Type        string `json:"type,omitempty"` // Host|Network|Range|FQDN|NetworkGroup
	Overridable bool   `json:"overridable,omitempty"`
}

type FMCNetworkObject struct {
	Id          string `json:"id,omitempty"`
	Name        string `json:"name,omitempty"`
	Type        string `json:"type,omitempty"` // Host|Network|Range|FQDN
	Value       string `json:"value,omitempty"`
	Overridable bool   `json:"overridable,omitempty"`
}

type fmcNetworkObjects struct {
	Items []*FMCNetworkObject `json:"items"`
}

type fmcNetworkObjectsPager struct {
	pager
}

func (p *fmcNetworkObjectsPager) Next() (
	out []*FMCNetworkObject,
	err error,
) {
	resp := &fmcNetworkObjects{}
	err = p.next(resp)
	if err == EOF {
		return
	}
	if err != nil {
		err = fmt.Errorf("fmc network objects pager: next: %v", err)
		return
	}
	return resp.Items, nil
}

type FMCNetworkGroupMetadata struct {
	Domain struct {
		Id   string `json:"id,omitempty"`
		Name string `json:"name,omitempty"`
		Type string `json:"type,omitempty"` // Domain
	}
}

type FMCNetworkLiteral struct {
	Type  string `json:"type"` // Network|Host
	Value string `json:"value"`
}

type FMCNetworkGroup struct {
	Id          string                   `json:"id,omitempty"`
	Name        string                   `json:"name,omitempty"`
	Type        string                   `json:"type,omitempty"`
	Overridable bool                     `json:"overridable,omitempty"`
	Metadata    *FMCNetworkGroupMetadata `json:"metadata,omitempty"`
	Literals    []*FMCNetworkLiteral     `json:"literals,omitempty"`
	Objects     []*FMCNetworkObjectRef   `json:"objects,omitempty"`
}

type fmcNetworkGroups struct {
	Items []*FMCNetworkGroup `json:"items"`
}

type fmcNetworkGroupsPager struct {
	pager
}

func (p *fmcNetworkGroupsPager) Next() (
	out []*FMCNetworkGroup,
	err error,
) {
	resp := &fmcNetworkGroups{}
	err = p.next(resp)
	if err == EOF {
		return
	}
	if err != nil {
		err = fmt.Errorf("fmc network groups pager: next: %v", err)
		return
	}
	return resp.Items, nil
}

type nameToUUID struct {
	Name string
	UUID string
}

func New(
	ctx context.Context,
	cl *http.Client,
	host,
	user,
	pass string,
) (*FMC, error) {
	f := &FMC{
		ctx:       ctx,
		client:    cl,
		host:      host,
		basePath:  "/api/fmc_config/v1",
		user:      user,
		pass:      pass,
		authMutex: &sync.Mutex{},
		domains:   make(map[string]string),
	}
	var err error
	if host == "" || user == "" || pass == "" {
		err = errors.New("Empty host, user, or pass")
	}
	return f, err
}

type FMC struct {
	ctx            context.Context
	client         *http.Client
	host           string
	basePath       string
	user           string
	pass           string
	authToken      string
	refreshToken   string
	refreshCount   int
	tokenExpiry    time.Time
	watchdogCancel context.CancelFunc
	authMutex      *sync.Mutex
	domains        map[string]string
}

func (f *FMC) authTokenInvalid() bool {
	if f.authToken == "" ||
		f.refreshCount > MaxAuthTokenRefreshes ||
		time.Until(f.tokenExpiry) <= time.Duration(0) {
		return true
	} else {
		return false
	}
}

func (f *FMC) resetRefreshToken() {
	f.refreshToken = ""
	f.refreshCount = 0
	f.tokenExpiry = time.Time{}
	return
}

func (f *FMC) processAuthResponse(
	resp *http.Response,
) error {
	/* TODO: Is this actually a failure? For the life of me, I
	can't recall why this was necessary. A successful
	authentication with status 204 is presented here:
	https://www.cisco.com/c/en/us/support/docs/security/firepower-management-center/215918-how-to-generate-authentication-token-for.html
	*/
	if resp.StatusCode != FMCStatusCodeNoContent {
		return newFMCError(resp, f.host, "authentication failure")
	}
	f.tokenExpiry = NewAuthTokenExpiry(time.Now())
	f.authToken = resp.Header.Get("X-Auth-Access-Token")
	f.refreshToken = resp.Header.Get("X-Auth-Refresh-Token")
	return nil
}

// See https://www.cisco.com/c/en/us/td/docs/security/firepower/620/api/REST/Firepower_Management_Center_REST_API_Quick_Start_Guide_620/Connecting_with_a_Client.html
func (f *FMC) Authenticate() error {
	f.authMutex.Lock()
	defer f.authMutex.Unlock()

	if f.authTokenInvalid() ||
		maxAuthTokenRefreshesHasBeenReached(f.refreshCount) ||
		authTokenRefreshDeadlineHasPassed(f.tokenExpiry) {
		if f.watchdogCancel != nil {
			f.watchdogCancel()
		}
		f.resetRefreshToken()

		req, err := makeAuthAccessRequest(
			f.ctx,
			f.host,
			f.user,
			f.pass,
		)
		if err != nil {
			return err
		}
		resp, err := f.client.Do(req)
		if err != nil {
			return err
		}
		var domains []nameToUUID
		json.Unmarshal(
			[]byte(resp.Header.Get("domains")),
			&domains,
		)
		for i := 0; i < len(domains); i++ {
			f.domains[domains[i].Name] = domains[i].UUID
		}
		if err = f.processAuthResponse(resp); err != nil {
			return err
		}
		info("authenticated to FMC")

		ctx, cancel := context.WithCancel(f.ctx)
		f.watchdogCancel = cancel
		f.startTokenWatchdog(ctx)
	} else {
		if !withinAuthTokenRefreshWindow(f.tokenExpiry) {
			return nil
		}
		req, err := makeAuthRefreshRequest(
			f.ctx,
			f.host,
			f.authToken,
			f.refreshToken,
		)
		if err != nil {
			return err
		}
		resp, err := f.client.Do(req)
		if err != nil {
			return err
		}
		if err = f.processAuthResponse(resp); err != nil {
			return err
		}
		info("refreshed FMC auth token")

		f.refreshCount++
	}
	return nil
}

func (f *FMC) startTokenWatchdog(ctx context.Context) {
	info("starting auth token watchdog")
	go func() {
		defer info("stopping auth token watchdog")
		for {
			select {
			case <-ctx.Done():
				return
			case <-time.After(100 * time.Millisecond):
				if f.authTokenInvalid() {
					info("auth token has invalidated")
					return
				}
				if err := f.Authenticate(); err != nil {
					warn(err.Error())
					return
				}
			}
		}
	}()
	return
}

func (f *FMC) get(url string) (
	resp *http.Response,
	err error,
) {
	req, err :=
		http.NewRequestWithContext(f.ctx, "GET", url, nil)
	if err != nil {
		return
	}
	req.Header.Add("X-Auth-Access-Token", f.authToken)
	resp, err = f.client.Do(req)
	return
}

func (f *FMC) post(
	url string,
	body io.Reader,
) (resp *http.Response, err error) {
	req, err :=
		http.NewRequestWithContext(f.ctx, "POST", url, body)
	if err != nil {
		return
	}
	req.Header.Add("Content-Type", "application/json")
	req.Header.Add("X-Auth-Access-Token", f.authToken)
	resp, err = f.client.Do(req)
	return
}

func (f *FMC) Domain(s string) (string, bool) {
	v, ok := f.domains[s]
	return v, ok
}

func (f *FMC) Domains() []string {
	keys := make([]string, 0, len(f.domains))
	for k := range f.domains {
		keys = append(keys, k)
	}
	return keys
}

func (f *FMC) AccessPolicies(
	domain string,
	pageSize uint64,
) (*fmcAccessPoliciesPager, error) {
	domainUUID, ok := f.domains[domain]
	if !ok {
		return nil, fmt.Errorf("access policies: unknown domain")
	}
	url := fmt.Sprintf(
		"https://%s%s/domain/%s/policy/accesspolicies",
		f.host,
		f.basePath,
		domainUUID,
	)
	return &fmcAccessPoliciesPager{
		pager: newPager(url, f, pageSize),
	}, nil
}

func (f *FMC) AccessPolicyRules(
	domain,
	policyUUID string,
	pageSize uint64,
) (*fmcAccessPolicyRulesPager, error) {
	domainUUID, ok := f.domains[domain]
	if !ok {
		return nil, fmt.Errorf("access policy rules: unknown domain")
	}
	url := fmt.Sprintf(
		"https://%s%s/domain/%s/policy/accesspolicies/%s/accessrules",
		f.host,
		f.basePath,
		domainUUID,
		policyUUID,
	)
	return &fmcAccessPolicyRulesPager{
		pager: newPager(url, f, pageSize),
	}, nil
}

func (f *FMC) AddAccessPolicyRule(
	domain,
	policyUUID string,
	rule FMCAccessPolicyRule,
) error {
	domainUUID, ok := f.domains[domain]
	if !ok {
		return fmt.Errorf("add access policy rule: unknown domain")
	}
	category := url.PathEscape(rule.Metadata.Category)
	url := fmt.Sprintf(
		"https://%s%s/domain/%s/policy/accesspolicies/%s/accessrules?bulk=true&category=%s",
		f.host,
		f.basePath,
		domainUUID,
		policyUUID,
		category,
	)
	rule.Metadata = nil
	//rule.SourceZones.Objects[0].Name = ""
	//rule.DestinationZones.Objects[0].Name = ""
	reqBody, err := json.Marshal(rule)
	if err != nil {
		return fmt.Errorf("add access policy rule: failed to marshal json for rule: %v: '%#v'", err, rule)
	}
	if len(rule.SourceNetworks.Objects) > 42 &&
		len(rule.SourceNetworks.Objects) < 50 ||
		len(rule.DestinationNetworks.Objects) > 42 &&
			len(rule.DestinationNetworks.Objects) < 50 {
		log.Printf("WARNING: network object count is approaching limit of 50: '%s'", reqBody)
	}
	if len(rule.SourceNetworks.Objects) == 50 ||
		len(rule.DestinationNetworks.Objects) == 50 {
		log.Printf("WARNING: limit of 50 network objects reached: '%s'", reqBody)
	}
	if len(rule.SourceNetworks.Objects) > 50 ||
		len(rule.DestinationNetworks.Objects) > 50 {
		return fmt.Errorf("add access policy rule: unable to add rule: network object count exceeds 50-object limit: '%s'", reqBody)
	}
	resp, err := f.post(url, bytes.NewReader(reqBody))
	if err != nil {
		return fmt.Errorf("add access policy rule: failed to post rule: %v: '%s'", err, reqBody)
	}
	defer resp.Body.Close()
	respBody, err := io.ReadAll(resp.Body)
	if err != nil {
		log.Printf("add access policy rule: failed to read response body: %v", err)
	}
	/* According to https://www.cisco.com/c/en/us/td/docs/security/firepower/620/api/REST/Firepower_Management_Center_REST_API_Quick_Start_Guide_620/Objects_in_the_REST_API.html
	the API cannot accept a message with a payload greater
	than 20480 bytes (20 KB). This applies to both the REST
	API and to API Explorer. This is not a configurable
	parameter.  If a message exceeds this limit, the API will
	give an HTTP 422 error.

	By contrast, https://www.cisco.com/c/en/us/td/docs/security/firepower/70/api/REST/firepower_management_center_rest_api_quick_start_guide_70/Objects_In_The_REST_API.html
	claims the API cannot accept a message with payload
	greater than 2048000 bytes (2 MB). I can only hope Cisco
	typo-ed the former when they meant the latter. */
	if resp.StatusCode == FMCStatusCodeUnprocessableEntity {
		log.Printf("DEBUG: req body: %s", "["+string(reqBody)+"]")
		return fmt.Errorf("add access policy rule: server returned error: %+v: %s", resp, string(respBody))
	}
	/* From https://www.cisco.com/c/en/us/td/docs/security/firepower/620/api/REST/Firepower_Management_Center_REST_API_Quick_Start_Guide_620/Objects_in_the_REST_API.html:

	> The FMC REST API implements rate limiting to reduce
	> network load.
	>
	> The API will accept no more than 120 messages per minute
	> from an individual IP address. It will only allow 10
	> simultaneous connections *per IP address*. These are not
	> configurable parameters.
	>
	> If a client exceeds these limits, the API will give an
	> HTTP 429 error.

	However, https://www.cisco.com/c/en/us/td/docs/security/firepower/70/api/REST/firepower_management_center_rest_api_quick_start_guide_70/Objects_In_The_REST_API.html
	states the following:

	> – "Too Many Requests". Too many requests were sent to
	> the API. This error will occur if you send more than 120
	> requests per minute.
	> – Too many concurrent requests.  The system cannot
	> accept more than 10 parallel requests *from all
	> clients*.
	> – Too many write operations per server. The API will
	> only allow one PUT, POST, or DELETE request *per user*
	> on a server at a time.

	If true, these latter limits constitute a substantial
	reduction in acceptable client behaviors.

	TODO: Prevent this occurring in the first place. */
	if resp.StatusCode == FMCStatusCodeTooManyRequests {
		log.Printf("DEBUG: req body: %s", "["+string(reqBody)+"]")
		return fmt.Errorf("add access policy rule: server returned error: %+v: %s", resp, string(respBody))
	}
	if resp.StatusCode != FMCStatusCodeAccepted {
		log.Printf("DEBUG: req body: %s", "["+string(reqBody)+"]")
		return fmt.Errorf("add access policy rule: server returned error: %+v: %s", resp, string(respBody))
	}
	//log.Printf("DEBUG: post response body: %s", string(respBody))
	return nil
}

func (f *FMC) NetworkHostObjects(
	domain string,
	pageSize uint64,
) (*fmcNetworkObjectsPager, error) {
	domainUUID, ok := f.domains[domain]
	if !ok {
		return nil, fmt.Errorf("network host objects: unknown domain")
	}
	url := fmt.Sprintf(
		"https://%s%s/domain/%s/object/hosts",
		f.host,
		f.basePath,
		domainUUID,
	)
	return &fmcNetworkObjectsPager{
		pager: newPager(url, f, pageSize),
	}, nil
}

/* Retrieve both network and host objects
 */
func (f *FMC) NetworkAddresses(
	domain string,
	pageSize uint64,
) (*fmcNetworkObjectsPager, error) {
	domainUUID, ok := f.domains[domain]
	if !ok {
		return nil, fmt.Errorf("network addresses: unknown domain")
	}
	url := fmt.Sprintf(
		"https://%s%s/domain/%s/object/networkaddresses",
		f.host,
		f.basePath,
		domainUUID,
	)
	return &fmcNetworkObjectsPager{
		pager: newPager(url, f, pageSize),
	}, nil
}

/* Retrieve network groups
 */
func (f *FMC) NetworkGroups(
	domain string,
	pageSize uint64,
) (*fmcNetworkGroupsPager, error) {
	domainUUID, ok := f.domains[domain]
	if !ok {
		return nil, fmt.Errorf("network groups: unknown domain")
	}
	url := fmt.Sprintf(
		"https://%s%s/domain/%s/object/networkgroups",
		f.host,
		f.basePath,
		domainUUID,
	)
	return &fmcNetworkGroupsPager{
		pager: newPager(url, f, pageSize),
	}, nil
}

// See https://www.cisco.com/c/en/us/td/docs/security/firepower/620/api/REST/Firepower_Management_Center_REST_API_Quick_Start_Guide_620/Connecting_with_a_Client.html
func makeAuthAccessRequest(
	ctx context.Context,
	host,
	user,
	pass string,
) (*http.Request, error) {
	url := fmt.Sprintf(
		"https://%s/api/fmc_platform/v1/auth/generatetoken",
		host,
	)
	req, err :=
		http.NewRequestWithContext(ctx, "POST", url, nil)
	if err != nil {
		return req, err
	}
	req.SetBasicAuth(user, pass)
	return req, nil
}

// See https://www.cisco.com/c/en/us/td/docs/security/firepower/620/api/REST/Firepower_Management_Center_REST_API_Quick_Start_Guide_620/Connecting_with_a_Client.html
func makeAuthRefreshRequest(
	ctx context.Context,
	host,
	authToken,
	refreshToken string,
) (*http.Request, error) {
	url := fmt.Sprintf(
		"https://%s/api/fmc_platform/v1/auth/refreshtoken",
		host,
	)
	req, err :=
		http.NewRequestWithContext(ctx, "POST", url, nil)
	if err != nil {
		return req, err
	}
	req.Header.Add("X-Auth-Access-Token", authToken)
	req.Header.Add("X-Auth-Refresh-Token", refreshToken)
	return req, nil
}

func warn(m string) {
	log.Printf("Warn: %s\n", m)
	return
}

func info(m string) {
	log.Printf("info: %s\n", m)
	return
}

func debug(m string) {
	log.Printf("debug: %s\n", m)
	return
}
